/*
 * Copyright 2021 Arne Vajhøj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.nettool.loadtest;

import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;
import javax.swing.border.EmptyBorder;

public class LoadTester extends JFrame {
	private static JPanel createGeneralPanel(int row, int col) {
        JPanel res = new JPanel();
        res.setBorder(new EmptyBorder(5, 5, 5, 5));
        res.setLayout(new GridLayout(row, col, 5, 5));
        return res;
	}
	private static JPanel createPair(JComponent comp1, JComponent comp2) {
		JPanel res = new JPanel();
        res.setLayout(new GridLayout(1, 2, 5, 5));
        res.add(comp1);
        res.add(comp2);
		return res;
	}
	private static JPanel createPair(String lbl, JComponent comp) {
		return createPair(new JLabel(lbl), comp);
	}
    private static final long serialVersionUID = 1L;
    private Object lock = new Object();
    private JTextField url;
    private JTextField nothreads;
    private JTextField reqperthr;
    private JComboBox<String> keepalive;
    private JButton start;
    private JButton stop;
    private JButton showlasterr;
    private JLabel status;
    private JLabel good;
    private JLabel bad;
    private JLabel avgtim;
    private JLabel maxtim;
    private JLabel throughput;
    private int ngood;
    private int nbad;
    private int sumtim;
    private int curmaxtim;
    private ExecutorService es;
    private int waiting;
    private String lasterr;
    public LoadTester() {
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        setTitle("Load tester");
        getContentPane().setLayout(new GridLayout(2, 1, 5, 5));
        // upper half
        JPanel upper = createGeneralPanel(6, 1);
        url = new JTextField("http://localhost/");
        upper.add(createPair("URL:", url));
        keepalive = new JComboBox<String>(new String[] { "Yes", "No" });
        keepalive.setSelectedIndex(0);
        upper.add(createPair("Keep alive:", keepalive));
        nothreads = new JTextField("100");
        upper.add(createPair("Number threads:", nothreads));
        reqperthr = new JTextField("100");
        upper.add(createPair("Requests per thread:", reqperthr));
        start = new JButton("Start");
        start.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                final int nt = Integer.parseInt(nothreads.getText());
                final int rpt = Integer.parseInt(reqperthr.getText());
                es = Executors.newFixedThreadPool(nt);
                waiting = rpt * nt;
                ngood = 0;
                nbad = 0;
                sumtim = 0;
                curmaxtim = 0;
                lasterr = "";
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        status.setText("Running");
                        start.setEnabled(false);
                        stop.setEnabled(true);
                    }
                });
                final long total_t1 = System.currentTimeMillis();
                for(int i = 0; i < rpt * nt; i++) {
                    es.submit(new Runnable() {
                        public void run() {
                            long t1 = System.currentTimeMillis();
                            try {
                                HttpURLConnection con = (HttpURLConnection) (new URL(url.getText())).openConnection();
                                if(keepalive.getSelectedIndex() == 1) {
                                	con.addRequestProperty("Connection", "Close");
                                }
                                if(con.getResponseCode() == HttpURLConnection.HTTP_OK) {
                                    InputStream is = con.getInputStream();
                                    byte[] b = new byte[10000];
                                    while(is.read(b) > 0);
                                    is.close();
                                    synchronized(lock) {
                                        ngood++;
                                    }
                                } else {
                                    synchronized(lock) {
                                    	lasterr = con.getResponseCode() + ": " + con.getResponseMessage();
                                        showlasterr.setEnabled(true);
                                        nbad++;
                                    }
                                }
                                con.disconnect();
                            } catch (MalformedURLException e) {
                                synchronized(lock) {
                                	lasterr = e.getMessage();
                                    showlasterr.setEnabled(true);
                                    nbad++;
                                }
                            } catch (IOException e) {
                                synchronized(lock) {
                                	lasterr = e.getMessage();
                                    showlasterr.setEnabled(true);
                                    nbad++;
                                }
                            }
                            long t2 = System.currentTimeMillis();
                            synchronized(lock) {
                                sumtim += (t2 - t1);
                                curmaxtim = Math.max(curmaxtim, (int)(t2 - t1));
                                waiting--;
                                if(waiting <= 0) {
                                    final long total_t2 = System.currentTimeMillis();
                                    SwingUtilities.invokeLater(new Runnable() {
                                        public void run() {
                                            status.setText("Done");;
                                            start.setEnabled(true);
                                            stop.setEnabled(false);
                                            if(lasterr.equals("")) showlasterr.setEnabled(false);
											good.setText(Integer.toString(ngood));
											bad.setText(Integer.toString(nbad));
											avgtim.setText(Integer.toString(sumtim / (rpt * nt)) + " ms");
											maxtim.setText(Integer.toString(curmaxtim) + " ms");
											throughput.setText(Long.toString(rpt * nt * 1000L / (total_t2 - total_t1)) + " req/sec");
                                        }
                                    });
                                }
                            }
                        }
                    });
                }
            }
        });
        stop = new JButton("Stop");
        stop.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                es.shutdownNow();
                waiting = 0;
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        status.setText("Aborted");
                        start.setEnabled(true);
                        stop.setEnabled(false);
                        if(lasterr.equals("")) showlasterr.setEnabled(false);
                    }
                });
            }
        });
        stop.setEnabled(false);
        upper.add(createPair(start, stop));
        showlasterr = new JButton("Show last error");
        showlasterr.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				JOptionPane.showMessageDialog(null, lasterr);
			}
        });
        showlasterr.setEnabled(false);
        upper.add(showlasterr);
        getContentPane().add(upper);
        // lower half
        JPanel lower = createGeneralPanel(6, 1);
        status = new JLabel("");
        lower.add(createPair("Status:", status));
        good = new JLabel("");
        lower.add(createPair("Good requests:", good));
        bad = new JLabel("");
        lower.add(createPair("Bad requests:", bad));
        avgtim = new JLabel("");
        lower.add(createPair("Average time:", avgtim));
        maxtim = new JLabel("");
        lower.add(createPair("Max time:", maxtim));
        throughput = new JLabel("");
        lower.add(createPair("Throughput:", throughput));
        getContentPane().add(lower);
        pack();
    }
    public static void main(String[] args) {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                JFrame f = new LoadTester();
                f.setVisible(true);
            }
        });
    }
} 
